PROGRAM gkvunit
!-------------------------------------------------------------------------------
!
!  Unit conversion for GKV
!
!    exp2gkv: Experimental unit -> GKV unit
!    gkv2exp: GKV unit -> Experimental unit
!
!-------------------------------------------------------------------------------
  implicit none
  character(len=1) :: conv_type

    write(*,'(a)') "Which conversion? Type 1 or 2."
    write(*,'(a)') "  1: Experimental unit -> GKV unit  (exp.in -> gkv.out)"
    write(*,'(a)') "  2: GKV unit -> Experimental unit  (gkv.in -> exp.out)"
    read(*,*) conv_type

    if (conv_type == "1") then
      write(*,'(a)') "### Exp -> GKV ###"
      call exp2gkv
    else if (conv_type == "2") then
      write(*,'(a)') "### GKV -> Exp ###"
      call gkv2exp
    else
      write(*,'(a)') "Invalid conversion type."
      stop
    end if

END PROGRAM gkvunit


MODULE parameters
!-------------------------------------------------------------------------------
!
!  Parameters
!
!-------------------------------------------------------------------------------
  implicit none

!- constants -
  real(kind=8), parameter :: pi = 4.d0 * atan(1.d0) ! Pi
  real(kind=8), parameter :: amu = 1.672621777d-27  ! Atomic mass unit [kg]
  real(kind=8), parameter :: ele = 1.602176565d-19  ! Elementary charge unit [C]
  real(kind=8), parameter :: ep0 = 8.854187817d-12
                                      ! Vacuum permittivity [m^-3*kg^-1*s^4*A^2]
  real(kind=8), parameter :: mu0 = pi * 4.d-7
                                      ! Vacuum permibility [m*kg*s^-2*A^-2]

!- parameters for exp.inout -
  integer :: expnml = 10
  real(kind=8) :: m_ref, e_ref, n_ref, T_ref, L_ref, B_ref
  real(kind=8) :: R_exp, a_exp, r2a_exp, B_exp, q_exp, s_exp
  character(len=99) :: sname
  real(kind=8) :: mass, charge, dens, temp, Ln, Lt
  real(kind=8) :: rho_ref, delta, v_ref, time_ref

!- parameters for gkv.inout -
  integer :: gkvnml = 20
  real(kind=8) :: r_minor, eps_r, q_0, s_hat, lambda_i, beta
  real(kind=8) :: R0_Ln, R0_Lt, nu, Anum, Znum, fcs, sgn, tau
  real(kind=8) :: lnlambda

  integer :: olog = 30

END MODULE parameters


SUBROUTINE exp2gkv
!-------------------------------------------------------------------------------
!
!    exp2gkv: Experimental unit -> GKV unit
!
!-------------------------------------------------------------------------------
  use parameters
  implicit none

  integer :: ios

  namelist /ref/    m_ref, e_ref, n_ref, T_ref, L_ref, B_ref
  namelist /config/ R_exp, a_exp, r2a_exp, B_exp, q_exp, s_exp
  namelist /species/ sname, mass, charge, dens, temp, Ln, Lt

    open(expnml, file="exp.in", status="old")
    open(gkvnml, file="gkv.out")
    read(expnml, nml=ref)
    rho_ref = sqrt(m_ref*amu * T_ref*1d3*ele) / (e_ref*ele * B_ref)
    delta = rho_ref / L_ref
    v_ref = sqrt(T_ref*1d3*ele / (m_ref*amu))
    time_ref = L_ref / v_ref
                              !- for debug -
                                write(*,'(a)') "# Constants"
                                write(*,"(9g15.7)") pi, amu, ele, ep0, mu0
                                write(*,'(a)') "# Reference values"
                                write(*,"(9g15.7)") m_ref, e_ref, n_ref, &
                                                    T_ref, L_ref, B_ref
                                write(*,"(9g15.7)") rho_ref, delta, &
                                                    v_ref, time_ref
                              !-------------
    write(gkvnml, '(a)') "&ref"
    write(gkvnml, '(a,g15.7)') " m_ref = ", m_ref
    write(gkvnml, '(a,g15.7)') " e_ref = ", e_ref
    write(gkvnml, '(a,g15.7)') " n_ref = ", n_ref
    write(gkvnml, '(a,g15.7)') " T_ref = ", T_ref
    write(gkvnml, '(a,g15.7)') " L_ref = ", L_ref
    write(gkvnml, '(a,g15.7)') " B_ref = ", B_ref
    write(gkvnml, '(a)') "&end"
    write(gkvnml, *)

    read(expnml, nml=config)
                              !- for debug -
                                write(*,'(a)') "# Configuration"
                                write(*,"(9g15.7)") R_exp, a_exp, r2a_exp, &
                                                    B_exp, q_exp, s_exp
                              !-------------
    r_minor = r2a_exp * a_exp / rho_ref
    eps_r = r2a_exp * a_exp / R_exp
    q_0 = q_exp
    s_hat = s_exp
    lambda_i = ep0 * T_ref*1d3*ele / (n_ref*1d19 * (e_ref*ele)**2) / rho_ref**2
    beta = mu0 * n_ref*1d19 * T_ref*1d3*ele / B_ref**2
    write(gkvnml, '(a)') "&config"
    write(gkvnml, '(a,g15.7)') " r_minor = ", r_minor
    write(gkvnml, '(a,g15.7)') "   eps_r = ", eps_r
    write(gkvnml, '(a,g15.7)') "     q_0 = ", q_0
    write(gkvnml, '(a,g15.7)') "   s_hat = ", s_hat
    write(gkvnml, '(a,g15.7)') "lambda_i = ", lambda_i
    write(gkvnml, '(a,g15.7)') "    beta = ", beta
    write(gkvnml, '(a)') "&end"
    write(gkvnml, *)

    do
      read(expnml, nml=species, iostat=ios)
      if (ios /= 0) exit
                              !- for debug -
                                write(*,'(a)') "# Species"
                                write(*,"(a,9g15.7)") "'"//trim(sname)//"'", &
                                                      mass, charge, dens, &
                                                      temp, Ln, Lt
                              !-------------
      R0_Ln = R_exp / Ln
      R0_Lt = R_exp / Lt
      lnlambda = 10.d0
      nu = n_ref*1d19 * (e_ref*ele)**4 * lnlambda * L_ref  &
         / (8.d0 * sqrt(2.d0) * pi * ep0**2 * (T_ref*1d3*ele)**2)
      Anum = mass*amu / (m_ref*amu)
      Znum = abs(charge*ele) / (e_ref*ele)
      fcs = abs(charge*ele * dens*1d19) / (e_ref*ele * n_ref*1d19)
      sgn = dsign(1.d0, charge)
      tau = temp*1d3*ele / (T_ref*1d3*ele)
      write(gkvnml, '(a)') "&species"
      write(gkvnml, '(2a)')      "sname = ", "'"//trim(sname)//"'"
      write(gkvnml, '(a,g15.7)') "R0_Ln = ", R0_Ln
      write(gkvnml, '(a,g15.7)') "R0_Lt = ", R0_Lt
      write(gkvnml, '(a,g15.7)') "   nu = ", nu
      write(gkvnml, '(a,g15.7)') " Anum = ", Anum
      write(gkvnml, '(a,g15.7)') " Znum = ", Znum
      write(gkvnml, '(a,g15.7)') "  fcs = ", fcs
      write(gkvnml, '(a,g15.7)') "  sgn = ", sgn
      write(gkvnml, '(a,g15.7)') "  tau = ", tau
      write(gkvnml, '(a)') "&end"
      write(gkvnml, *)
    end do

    close(expnml)
    close(gkvnml)


END SUBROUTINE exp2gkv


SUBROUTINE gkv2exp
!-------------------------------------------------------------------------------
!
!    gkv2exp: GKV unit -> Experimental unit
!
!-------------------------------------------------------------------------------
  use parameters
  implicit none

  integer :: ios

  namelist /ref/    m_ref, e_ref, n_ref, T_ref, L_ref, B_ref
  namelist /config/ r_minor, eps_r, q_0, s_hat, lambda_i, beta
  namelist /species/ sname, R0_Ln, R0_Lt, nu, Anum, Znum, fcs, sgn, tau

    open(gkvnml, file="gkv.in", status="old")
    open(expnml, file="exp.out")
    read(gkvnml, nml=ref)
    rho_ref = sqrt(m_ref*amu * T_ref*1d3*ele) / (e_ref*ele * B_ref)
    delta = rho_ref / L_ref
    v_ref = sqrt(T_ref*1d3*ele / (m_ref*amu))
    time_ref = L_ref / v_ref
                              !- for debug -
                                write(*,'(a)') "###############################"
                                write(*,'(a)') "### NOTE: B_exp = B_ref and "//&
                                               "r2a_exp = 0.5 are assumed!"
                                write(*,'(a)') "###############################"
                                write(*,*)
                                write(*,'(a)') "# Constants"
                                write(*,"(9g15.7)") pi, amu, ele, ep0, mu0
                                write(*,'(a)') "# Reference values"
                                write(*,"(9g15.7)") m_ref, e_ref, n_ref, &
                                                    T_ref, L_ref, B_ref
                                write(*,"(9g15.7)") rho_ref, delta, &
                                                    v_ref, time_ref
                              !-------------
    write(expnml, '(a)') "&ref"
    write(expnml, '(a,g15.7)') " m_ref = ", m_ref
    write(expnml, '(a,g15.7)') " e_ref = ", e_ref
    write(expnml, '(a,g15.7)') " n_ref = ", n_ref
    write(expnml, '(a,g15.7)') " T_ref = ", T_ref
    write(expnml, '(a,g15.7)') " L_ref = ", L_ref
    write(expnml, '(a,g15.7)') " B_ref = ", B_ref
    write(expnml, '(a)') "&end"
    write(expnml, *)

    read(gkvnml, nml=config)
                              !- for debug -
                                write(*,'(a)') "# Configuration"
                                write(*,"(9g15.7)") r_minor, eps_r, q_0, &
                                                    s_hat, lambda_i, beta
                              !-------------
    B_exp = B_ref   !!!NOTE: Assumption used in GKV
    r2a_exp = 0.5d0 !!!NOTE: Additional assumption
    a_exp = r_minor * rho_ref / r2a_exp
    !a_exp = 0.558d0
    !r2a_exp = r_minor * rho_ref / a_exp
    R_exp = r2a_exp * a_exp / eps_r
    q_exp = q_0
    s_exp = s_hat
    write(expnml, '(a)') "&config"
    write(expnml, '(a,g15.7)') "   R_exp = ", R_exp
    write(expnml, '(a,g15.7)') "   a_exp = ", a_exp
    write(expnml, '(a,g15.7)') " r2a_exp = ", r2a_exp
    write(expnml, '(a,g15.7)') "   B_exp = ", B_exp
    write(expnml, '(a,g15.7)') "   q_exp = ", q_exp
    write(expnml, '(a,g15.7)') "   s_exp = ", s_exp
    write(expnml, '(a)') "&end"
    write(expnml, *)

    do
      read(gkvnml, nml=species, iostat=ios)
      if (ios /= 0) exit
                              !- for debug -
                                write(*,'(a)') "# Species"
                                write(*,"(a,9g15.7)") "'"//trim(sname)//"'", &
                                                      R0_Ln, R0_Lt, nu, &
                                                      Anum, Znum, fcs, sgn, tau
                              !-------------
      mass = Anum * m_ref
      charge = sgn * Znum * e_ref
      dens = fcs * n_ref * e_ref / Znum
      temp = tau * T_ref
      Ln = R_exp / R0_Ln
      Lt = R_exp / R0_Lt
      write(expnml, '(a)') "&species"
      write(expnml, '(2a)')      " sname = ", "'"//trim(sname)//"'"
      write(expnml, '(a,g15.7)') "  mass = ", mass
      write(expnml, '(a,g15.7)') "charge = ", charge
      write(expnml, '(a,g15.7)') "  dens = ", dens
      write(expnml, '(a,g15.7)') "  temp = ", temp
      write(expnml, '(a,g15.7)') "    Ln = ", Ln
      write(expnml, '(a,g15.7)') "    Lt = ", Lt
      write(expnml, '(a)') "&end"
      write(expnml, *)
    end do

    close(expnml)
    close(gkvnml)


END SUBROUTINE gkv2exp
