MODULE out_momintky
!-------------------------------------------------------------------------------
!
!     Output moments in (x,y)
!                                                   (S. Maeyama, 11 June 2014)
!
!-------------------------------------------------------------------------------
  use diag_header
  implicit none

  private

  public momintky, fpintintky


 CONTAINS


SUBROUTINE momintky( loop )
!-------------------------------------------------------------------------------
!
!     Output field energy in (kx,ky) at loop
!                                                   (S. Maeyama, 24 Aug 2014)
!
!-------------------------------------------------------------------------------
  use diag_rb, only : rb_phi_gettime, rb_phi_loop, rb_Al_loop
  use diag_geom, only : grootg, cfsrf, fct_e_energy, fct_m_energy

  integer, intent(in) :: loop

  real(kind=DP) :: time
  complex(kind=DP), dimension(-nx:nx,0:global_ny,-global_nz:global_nz-1) :: phi, Al
  real(kind=DP), dimension(-nx:nx,0:global_ny) :: e_energy, m_energy
  real(kind=DP), dimension(0:global_ny) :: wke, wkm
  real(kind=DP) :: fct
  integer :: mx, my, iz

    call rb_phi_gettime( loop, time )
    call rb_phi_loop( loop, phi )
    call rb_Al_loop( loop, Al )

    e_energy(:,:) = 0._DP
    m_energy(:,:) = 0._DP
    do iz = -global_nz, global_nz-1
      fct = grootg(iz) / cfsrf      ! flux-surface average
      do my = 0, global_ny
        do mx = -nx, nx
          e_energy(mx,my) = e_energy(mx,my)  &
                          + fct * fct_e_energy(mx,my,iz) * abs(phi(mx,my,iz))**2
          m_energy(mx,my) = m_energy(mx,my)  &
                          + fct * fct_m_energy(mx,my,iz) * abs(Al(mx,my,iz))**2
        end do
      end do
    end do

    wke(:) = 0._DP
    wkm(:) = 0._DP
    my = 0
      do mx = 0, nx
        wke(my) = wke(my) + e_energy(mx,my)
        wkm(my) = wkm(my) + m_energy(mx,my)
      end do
    do my = 1, global_ny
      do mx = -nx, nx
        wke(my) = wke(my) + e_energy(mx,my)
        wkm(my) = wkm(my) + m_energy(mx,my)
      end do
    end do

    write( 98, "(999g17.7e3)" ) time, sum(wke), wke(:)
    write( 99, "(999g17.7e3)" ) time, sum(wkm), wkm(:)

END SUBROUTINE momintky


SUBROUTINE fpintintky( loop )
!-------------------------------------------------------------------------------
!
!     Output field-particle interaction terms in (kx,ky) at loop
!                                                   (S. Maeyama, 17 Sep 2014)
!
!-------------------------------------------------------------------------------
  use diag_rb, only : rb_phi_gettime, rb_phi_loop, rb_Al_loop, rb_mom_imomisloop, loop_phi_sta, loop_phi_end
  use diag_geom, only : grootg, cfsrf, sgn, Znum

  integer, intent(in) :: loop

  real(kind=DP) :: time
  complex(kind=DP), dimension(-nx:nx,0:global_ny,-global_nz:global_nz-1) :: phi, dAldt
  complex(kind=DP), dimension(-nx:nx,0:global_ny,-global_nz:global_nz-1,0:ns-1) :: dndt, upara
  complex(kind=DP), dimension(-nx:nx,0:global_ny,-global_nz:global_nz-1,-2:2) :: wk
  real(kind=DP), dimension(-nx:nx,0:global_ny,0:ns-1) :: fpint_es, fpint_em
  real(kind=DP), dimension(0:global_ny,0:ns-1) :: wke, wkm
  real(kind=DP) :: fct
  integer :: mx, my, iz, is

    call rb_phi_gettime( loop, time )
    call rb_phi_loop( loop, phi )
    do is = 0, ns-1
      call rb_mom_imomisloop( 1, is, loop, upara(:,:,:,is) )
    end do

    if (loop == loop_phi_sta(snum)) then
      fct = 1._DP / (2._DP * dtout_ptn)
      call rb_Al_loop( loop  , wk(:,:,:, 0) )
      call rb_Al_loop( loop+1, wk(:,:,:,+1) )
      call rb_Al_loop( loop+2, wk(:,:,:,+2) )
      do iz = -global_nz, global_nz-1
        do my = 0, global_ny
          do mx = -nx, nx
            dAldt(mx,my,iz) = fct * ( -3._DP * wk(mx,my,iz, 0)  &
                                      +4._DP * wk(mx,my,iz,+1)  &
                                      -        wk(mx,my,iz,+2) )
          end do
        end do
      end do
      do is = 0, ns-1
        call rb_mom_imomisloop( 0, is, loop  , wk(:,:,:, 0) )
        call rb_mom_imomisloop( 0, is, loop+1, wk(:,:,:,+1) )
        call rb_mom_imomisloop( 0, is, loop+2, wk(:,:,:,+2) )
        do iz = -global_nz, global_nz-1
          do my = 0, global_ny
            do mx = -nx, nx
              dndt(mx,my,iz,is) = fct * ( -3._DP * wk(mx,my,iz, 0)  &
                                          +4._DP * wk(mx,my,iz,+1)  &
                                          -        wk(mx,my,iz,+2) )
            end do
          end do
        end do
      end do
    else if (loop == loop_phi_sta(snum)+1) then
      fct = 1._DP / (6._DP * dtout_ptn)
      call rb_Al_loop( loop-1, wk(:,:,:,-1) )
      call rb_Al_loop( loop  , wk(:,:,:, 0) )
      call rb_Al_loop( loop+1, wk(:,:,:,+1) )
      call rb_Al_loop( loop+2, wk(:,:,:,+2) )
      do iz = -global_nz, global_nz-1
        do my = 0, global_ny
          do mx = -nx, nx
            dAldt(mx,my,iz) = fct * ( -2._DP * wk(mx,my,iz,-1)  &
                                      -3._DP * wk(mx,my,iz, 0)  &
                                      +6._DP * wk(mx,my,iz,+1)  &
                                      -        wk(mx,my,iz,+2) )
          end do
        end do
      end do
      do is = 0, ns-1
        call rb_mom_imomisloop( 0, is, loop-1, wk(:,:,:,-1) )
        call rb_mom_imomisloop( 0, is, loop  , wk(:,:,:, 0) )
        call rb_mom_imomisloop( 0, is, loop+1, wk(:,:,:,+1) )
        call rb_mom_imomisloop( 0, is, loop+2, wk(:,:,:,+2) )
        do iz = -global_nz, global_nz-1
          do my = 0, global_ny
            do mx = -nx, nx
              dndt(mx,my,iz,is) = fct * ( -2._DP * wk(mx,my,iz,-1)  &
                                          -3._DP * wk(mx,my,iz, 0)  &
                                          +6._DP * wk(mx,my,iz,+1)  &
                                          -        wk(mx,my,iz,+2) )
            end do
          end do
        end do
      end do
    else if (loop == loop_phi_end(enum)-1) then
      fct = 1._DP / (6._DP * dtout_ptn)
      call rb_Al_loop( loop-2, wk(:,:,:,-2) )
      call rb_Al_loop( loop-1, wk(:,:,:,-1) )
      call rb_Al_loop( loop  , wk(:,:,:, 0) )
      call rb_Al_loop( loop+1, wk(:,:,:,+1) )
      do iz = -global_nz, global_nz-1
        do my = 0, global_ny
          do mx = -nx, nx
            dAldt(mx,my,iz) = fct * ( +        wk(mx,my,iz,-2)  &
                                      -6._DP * wk(mx,my,iz,-1)  &
                                      +3._DP * wk(mx,my,iz, 0)  &
                                      +2._DP * wk(mx,my,iz,+1) )
          end do
        end do
      end do
      do is = 0, ns-1
        call rb_mom_imomisloop( 0, is, loop-2, wk(:,:,:,-2) )
        call rb_mom_imomisloop( 0, is, loop-1, wk(:,:,:,-1) )
        call rb_mom_imomisloop( 0, is, loop  , wk(:,:,:, 0) )
        call rb_mom_imomisloop( 0, is, loop+1, wk(:,:,:,+1) )
        do iz = -global_nz, global_nz-1
          do my = 0, global_ny
            do mx = -nx, nx
              dndt(mx,my,iz,is) = fct * ( +        wk(mx,my,iz,-2)  &
                                          -6._DP * wk(mx,my,iz,-1)  &
                                          +3._DP * wk(mx,my,iz, 0)  &
                                          +2._DP * wk(mx,my,iz,+1) )
            end do
          end do
        end do
      end do
    else if (loop == loop_phi_end(enum)) then
      fct = 1._DP / (2._DP * dtout_ptn)
      call rb_Al_loop( loop-2, wk(:,:,:,-2) )
      call rb_Al_loop( loop-1, wk(:,:,:,-1) )
      call rb_Al_loop( loop  , wk(:,:,:, 0) )
      do iz = -global_nz, global_nz-1
        do my = 0, global_ny
          do mx = -nx, nx
            dAldt(mx,my,iz) = fct * ( +        wk(mx,my,iz,-2)  &
                                      -4._DP * wk(mx,my,iz,-1)  &
                                      +3._DP * wk(mx,my,iz, 0) )
          end do
        end do
      end do
      do is = 0, ns-1
        call rb_mom_imomisloop( 0, is, loop-2, wk(:,:,:,-2) )
        call rb_mom_imomisloop( 0, is, loop-1, wk(:,:,:,-1) )
        call rb_mom_imomisloop( 0, is, loop  , wk(:,:,:, 0) )
        do iz = -global_nz, global_nz-1
          do my = 0, global_ny
            do mx = -nx, nx
              dndt(mx,my,iz,is) = fct * ( +        wk(mx,my,iz,-2)  &
                                          -4._DP * wk(mx,my,iz,-1)  &
                                          +3._DP * wk(mx,my,iz, 0) )
            end do
          end do
        end do
      end do
    else ! loop_phi_sta(snum)+2 <= loop <= loop_phi_end(enum)
      fct = 1._DP / (12._DP * dtout_ptn)
      call rb_Al_loop( loop-2, wk(:,:,:,-2) )
      call rb_Al_loop( loop-1, wk(:,:,:,-1) )
      call rb_Al_loop( loop+1, wk(:,:,:,+1) )
      call rb_Al_loop( loop+2, wk(:,:,:,+2) )
      do iz = -global_nz, global_nz-1
        do my = 0, global_ny
          do mx = -nx, nx
            dAldt(mx,my,iz) = fct * ( +        wk(mx,my,iz,-2)  &
                                      -8._DP * wk(mx,my,iz,-1)  &
                                      +8._DP * wk(mx,my,iz,+1)  &
                                      -        wk(mx,my,iz,+2) )
          end do
        end do
      end do
      do is = 0, ns-1
        call rb_mom_imomisloop( 0, is, loop-2, wk(:,:,:,-2) )
        call rb_mom_imomisloop( 0, is, loop-1, wk(:,:,:,-1) )
        call rb_mom_imomisloop( 0, is, loop+1, wk(:,:,:,+1) )
        call rb_mom_imomisloop( 0, is, loop+2, wk(:,:,:,+2) )
        do iz = -global_nz, global_nz-1
          do my = 0, global_ny
            do mx = -nx, nx
              dndt(mx,my,iz,is) = fct * ( +        wk(mx,my,iz,-2)  &
                                          -8._DP * wk(mx,my,iz,-1)  &
                                          +8._DP * wk(mx,my,iz,+1)  &
                                          -        wk(mx,my,iz,+2) )
            end do
          end do
        end do
      end do
    end if

    fpint_es(:,:,:) = 0._DP
    fpint_em(:,:,:) = 0._DP
    do is = 0, ns-1
      do iz = -global_nz, global_nz-1
        fct = grootg(iz) / cfsrf      ! flux-surface average
        do my = 0, global_ny
          do mx = -nx, nx
            fpint_es(mx,my,is) = fpint_es(mx,my,is) + fct  &
                               * real( - conjg( phi(mx,my,iz) ) * sgn(is) * Znum(is) &
                                       * dndt(mx,my,iz,is), kind=DP )
            fpint_em(mx,my,is) = fpint_em(mx,my,is) + fct  &
                               * real( - conjg( upara(mx,my,iz,is) ) * sgn(is) * Znum(is) &
                                       * dAldt(mx,my,iz), kind=DP )
          end do
        end do
      end do
    end do

    wke(:,:) = 0._DP
    wkm(:,:) = 0._DP
    do is = 0, ns-1
      my = 0
        do mx = 0, nx
          wke(my,is) = wke(my,is) + 2._DP * fpint_es(mx,my,is)
          wkm(my,is) = wkm(my,is) + 2._DP * fpint_em(mx,my,is)
        end do
      do my = 1, global_ny
        do mx = -nx, nx
          wke(my,is) = wke(my,is) + 2._DP * fpint_es(mx,my,is)
          wkm(my,is) = wkm(my,is) + 2._DP * fpint_em(mx,my,is)
        end do
      end do
    end do

    is = 0
    write( 96, "(999g17.7e3)" ) time, sum(wke(:,is)), wke(:,is)
    write( 97, "(999g17.7e3)" ) time, sum(wkm(:,is)), wkm(:,is)
    is = 1
    write( 98, "(999g17.7e3)" ) time, sum(wke(:,is)), wke(:,is)
    write( 99, "(999g17.7e3)" ) time, sum(wkm(:,is)), wkm(:,is)

END SUBROUTINE fpintintky


END MODULE out_momintky
